import ast
from unittest import mock

import pytest

from rattr.analyser.context import (
    Builtin,
    Class,
    Context,
    Func,
    Import,
    Name,
    RootContext,
    SymbolTable,
)


class TestContext:
    def test_add(self):
        # Test add
        root = Context(None)
        root.add(Name("var_one"))

        child = Context(root)
        child.add(Name("var_two"))

        assert "var_one" in root
        assert "var_one" in child

        assert "var_two" not in root
        assert "var_two" in child

        # Test add argument
        root = Context(None)
        root.add(Name("x"))

        fn_ctx = Context(root)
        fn_ctx.add(Name("x"), is_argument=True)

        assert root.symbol_table == SymbolTable(**{"x": Name("x")})
        assert fn_ctx.symbol_table == SymbolTable(**{"x": Name("x")})

    def test_add_all(self):
        root = Context(None)
        root.add_all([Name("a"), Name("b")])

        assert root.symbol_table == SymbolTable(
            **{
                "a": Name("a"),
                "b": Name("b"),
            }
        )

    def test_remove(self):
        # Remove from self
        root = Context(None)
        root.add_all([Name("a"), Name("b")])
        assert root.symbol_table == SymbolTable(
            **{
                "a": Name("a"),
                "b": Name("b"),
            }
        )

        root.remove("a")
        assert root.symbol_table == SymbolTable(**{"b": Name("b")})

        # Remove from parent
        root = Context(None)
        root.add_all([Name("a"), Name("b")])
        child = Context(root)

        assert root.symbol_table == SymbolTable(
            **{
                "a": Name("a"),
                "b": Name("b"),
            }
        )
        assert child.symbol_table == SymbolTable()

        child.remove("a")
        assert root.symbol_table == SymbolTable(**{"b": Name("b")})
        assert child.symbol_table == SymbolTable()

        # Remove non-existent
        root = Context(None)
        root.add_all([Name("a"), Name("b")])
        child = Context(root)

        root.remove("fake")
        child.remove("false")

        assert root.symbol_table == SymbolTable(
            **{
                "a": Name("a"),
                "b": Name("b"),
            }
        )
        assert child.symbol_table == SymbolTable()

    def test_remove_all(self):
        # Remove from self
        root = Context(None)
        root.add_all([Name("a"), Name("b")])
        assert root.symbol_table == SymbolTable(
            **{
                "a": Name("a"),
                "b": Name("b"),
            }
        )

        root.remove_all(list(root.symbol_table.names()))
        assert root.symbol_table == SymbolTable()

        # Mixed removal
        root = Context(None)
        root.add(Name("a"))
        child = Context(root)
        child.add(Name("b"))

        assert root.symbol_table == SymbolTable(**{"a": Name("a")})
        assert child.symbol_table == SymbolTable(**{"b": Name("b")})

        child.remove_all(["a", "b", "c"])

        assert root.symbol_table == SymbolTable()
        assert child.symbol_table == SymbolTable()

    def test_get(self):
        root = Context(None)
        root.add_all([Name("a"), Name("b")])

        assert root.get("a") == Name("a")
        assert root.get("b") == Name("b")
        assert root.get("x") is None

    def test_get_call_target(self, capfd, config):
        # No target
        root = Context(None)
        assert root.get_call_target("anything()", None) is None

        output, _ = capfd.readouterr()
        assert "unable to resolve call" in output

        # Target is not "Callable"
        root = Context(None)
        root.add(Name("a"))
        child = Context(root)

        assert child.get_call_target("a()", None) == Name("a")

        output, _ = capfd.readouterr()
        assert "is not callable" in output

        # Target is not "Callable" -- higher-order function
        root = Context(None)
        root.add(Name("a"))
        child = Context(root)

        root.get_call_target("a()", None)

        output, _ = capfd.readouterr()
        assert "likely a procedural parameter" in output

        # Target is not "Callable" -- higher-order function
        root = Context(None)
        root.add(Name("a"))
        child = Context(root)

        with config("show_low_priority_warnings", True):
            root.get_call_target("a.method()", None)

        output, _ = capfd.readouterr()
        assert "unable to resolve call to method" in output

        # Found target
        root = Context(None)
        a_symbol = Func("a", [], None, None)
        root.add(a_symbol)

        assert root.get_call_target("a()", None) == a_symbol

    def test_declares(self):
        root = Context(None)
        root.add(Name("a"))
        child = Context(root)
        child.add(Name("b"))

        assert root.declares("a")
        assert not root.declares("b")

        assert not child.declares("a")
        assert child.declares("b")

    def test_in(self):
        root = Context(None)
        root.add(Name("a"))
        child = Context(root)
        child.add(Name("b"))

        assert "a" in root
        assert "b" not in root

        assert "a" in child
        assert "b" in child

    def test_add_identifiers_to_context(self):
        # Single identifier
        expr = ast.parse("a = 1").body[0].targets[0]  # = a

        root = Context(None)

        assert root.symbol_table == SymbolTable()
        root.add_identifiers_to_context(expr)
        assert root.symbol_table == SymbolTable(**{"a": Name("a")})

        # Multiple identifiers per call (i.e. tuple assignment)
        expr = ast.parse("a, b, c = 1, 2, 3").body[0].targets[0]  # = (a, b, c)

        root = Context(None)

        assert root.symbol_table == SymbolTable()
        root.add_identifiers_to_context(expr)
        assert root.symbol_table == SymbolTable(
            **{
                "a": Name("a"),
                "b": Name("b"),
                "c": Name("c"),
            }
        )

    def test_del_identifiers_from_context(self):
        # Single identifier
        expr = ast.parse("del a").body[0].targets[0]  # = a

        root = Context(None)
        root.add(Name("a"))

        assert root.symbol_table == SymbolTable(**{"a": Name("a")})
        root.del_identifiers_from_context(expr)
        assert root.symbol_table == SymbolTable()

        # Multiple identifiers
        # `del, x, y, ...z` produces list of `ast.Name`s not a tuple
        expr_a = ast.parse("del a, b").body[0].targets[0]  # = a
        expr_b = ast.parse("del a, b").body[0].targets[1]  # = b

        root = Context(None)
        root.add(Name("a"))
        root.add(Name("b"))

        assert root.symbol_table == SymbolTable(
            **{
                "a": Name("a"),
                "b": Name("b"),
            }
        )
        root.del_identifiers_from_context(expr_a)
        root.del_identifiers_from_context(expr_b)
        assert root.symbol_table == SymbolTable()

    def test_push_arguments_to_context(self, parse):
        # No args
        arguments = (
            parse(
                """
            def fn():
                pass
        """
            )
            .body[0]
            .args
        )

        root = Context(None)

        assert root.symbol_table == SymbolTable()
        root.push_arguments_to_context(arguments)
        assert root.symbol_table == SymbolTable()

        # Many args
        arguments = (
            parse(
                """
            def fn(a, b, c=1, d=2, *x, **y):
                pass
        """
            )
            .body[0]
            .args
        )

        root = Context(None)

        assert root.symbol_table == SymbolTable()
        root.push_arguments_to_context(arguments)
        assert root.symbol_table == SymbolTable(
            **{
                "a": Name("a"),
                "b": Name("b"),
                "c": Name("c"),
                "d": Name("d"),
                "x": Name("x"),
                "y": Name("y"),
            }
        )

    def test_redefinition_in_scope(self):
        root = Context(None)
        root.add(Name("var_one"))
        root.add(Name("var_one"))

        assert root.symbol_table == SymbolTable(**{"var_one": Name("var_one")})

    def test_redefinition_in_child_scope(self):
        root = Context(None)
        root.add(Name("var_one"))

        child = Context(root)
        child.add(Name("var_one"))

        assert root.symbol_table == SymbolTable(**{"var_one": Name("var_one")})
        assert child.symbol_table == SymbolTable()

    def test_is_import(self):
        root = Context(None)
        root.add(Import("math", "math"))
        root.add(Import("np", "numpy"))
        root.add(Import("join", "os.path.join"))

        child = Context(root)

        # From within root
        assert root.is_import("math")
        assert root.is_import("np")
        assert root.is_import("join")
        assert not root.is_import("numpy")
        assert not root.is_import("os")
        assert not root.is_import("os.path")
        assert not root.is_import("os.path.join")

        # From within child
        assert child.is_import("math")
        assert child.is_import("np")
        assert child.is_import("join")
        assert not child.is_import("numpy")
        assert not child.is_import("os")
        assert not child.is_import("os.path")
        assert not child.is_import("os.path.join")


class TestRootContext_ContextFromPython:
    def test_module_level_attributes(self, parse):
        _ast = parse(
            """
            # A blank context
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None

        module_level_attributes = [
            "__annotations__",
            "__builtins__",
            "__cached__",
            "__doc__",
            "__file__",
            "__loader__",
            "__name__",
            "__package__",
            "__spec__",
        ]

        for attr in module_level_attributes:
            assert attr in _ctx.symbol_table

    def test_builtins(self, parse):
        _ast = parse(
            """
            # A blank context
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None

        # Some selected examples
        assert "print" in _ctx.symbol_table
        assert "max" in _ctx.symbol_table
        assert "len" in _ctx.symbol_table

        assert "Exception" in _ctx.symbol_table
        assert "KeyError" in _ctx.symbol_table
        assert "TypeError" in _ctx.symbol_table

    def test_affective_builtins(self, parse):
        _ast = parse(
            """
            # A blank context
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None

        assert _ctx.symbol_table.get("does_not_exist") is None
        assert _ctx.symbol_table.get("print") == Builtin("print", has_affect=False)
        assert _ctx.symbol_table.get("setattr") == Builtin("setattr", has_affect=True)


class TestRootContext_Imports:
    def test_import(self, parse, RootSymbolTable):
        _ast = parse(
            """
            import math
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Import("math", "math"),
        )

    def test_import_as(self, parse, RootSymbolTable):
        _ast = parse(
            """
            import math as m
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Import("m", "math"),
        )

    def test_import_list(self, parse, RootSymbolTable, capfd, config):
        _ast = parse(
            """
            import os, math
        """
        )
        with config("show_low_priority_warnings", True):
            _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Import("os"),
            Import("math"),
        )

        output, _ = capfd.readouterr()
        assert "info" in output

    def test_from_import(self, parse, RootSymbolTable):
        _ast = parse(
            """
            from os.path import isfile
            from math import sin, cos
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Import("isfile", "os.path.isfile"),
            Import("sin", "math.sin"),
            Import("cos", "math.cos"),
        )

    def test_from_import_as(self, parse, RootSymbolTable):
        _ast = parse(
            """
            from os import path as path_utils
            from math import power as exp, some_func
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Import("path_utils", "os.path"),
            Import("exp", "math.power"),
            Import("some_func", "math.some_func"),
        )

    def test_from_import_star(self, parse, RootSymbolTable, capfd):
        _ast = parse(
            """
            from math import *
        """
        )

        _ctx = RootContext(_ast)

        # Were error.fatal not to be called,
        # assert that it would have been correct
        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(Import("*", "math"))

        output, _ = capfd.readouterr()
        assert "warning" in output and "*" in output

    def test_nested_from_import(self, parse, RootSymbolTable):
        _ast = parse(
            """
            from os.path import join
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(Import("join", "os.path.join"))

    def test_multiple_starred_imports(self, parse, RootSymbolTable):
        _ast = parse(
            """
            from os import *
            from math import *
        """
        )

        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert Import("*", "os") in _ctx.symbol_table.symbols()
        assert Import("*", "math") in _ctx.symbol_table.symbols()

    def test_import_in_child_context(self, parse, RootSymbolTable):
        # In if, try, for, etc
        _ast = parse(
            """
            if something:
                from math import *
        """
        )

        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Import("*", "math"),
        )

        # In function, class, etc
        _ast = parse(
            """
            def fn():
                from math import *
        """
        )

        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn", [], None, None),
        )
        assert Import("*", "math") not in _ctx.symbol_table.symbols()


class TestRootContext_Assigns:
    def test_assignment(self, parse, RootSymbolTable):
        _ast = parse(
            """
            a = 42
            a = "reassigned!"

            b = hhgttg()

            x, y, = 1, 2
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Name("a"),
            Name("b"),
            Name("x"),
            Name("y"),
        )

    def test_ann_assignment(self, parse, RootSymbolTable):
        _ast = parse(
            """
            a: int = 42
            b: str = "an string!"
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Name("a"),
            Name("b"),
        )

    def test_aug_assignment(self, parse, RootSymbolTable):
        _ast = parse(
            """
            a = 7
            a += 5

            b %= "fmt"
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Name("a"),
            Name("b"),
        )

    @pytest.mark.py_3_8_plus()
    def test_walrus(self, parse, RootSymbolTable):
        _ast = parse(
            """
            a = (b := 5)
            x = (y, z := 10)
            """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Name("a"),
            Name("b"),
            Name("x"),
            Name("z"),
        )

    @pytest.mark.py_3_8_plus()
    def test_walrus_lambda(self, parse, RootSymbolTable):
        # Double lambda
        _ast = parse(
            """
            outer_lambda = (inner_lambda := lambda *a, **k: 0)
            """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("inner_lambda", [], "a", "k", False, None),
            Func("outer_lambda", [], "a", "k", False, None),
        )

        # Faux double lambda
        _ast = parse(
            """
            outer_lambda = (a, inner_lambda := lambda *a, **k: 0)
            """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("inner_lambda", [], "a", "k", False, None),
            Name("outer_lambda"),
        )


class TestRootContext_FunctionDefs:
    def test_function_def(self, parse, RootSymbolTable):
        _ast = parse(
            """
            def fn_one(a, b):
                pass

            def fn_two(c):
                var = "i shouldn't be in the root context!"
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn_one", ["a", "b"], None, None),
            Func("fn_two", ["c"], None, None),
        )

    def test_async_function_def(self, parse, RootSymbolTable):
        _ast = parse(
            """
            async def fn_one(a, b):
                pass

            async def fn_two(c):
                var = "i shouldn't be in the root context!"
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn_one", ["a", "b"], None, None, is_async=True),
            Func("fn_two", ["c"], None, None, is_async=True),
        )

    def test_lambda_anonymous(self, parse, RootSymbolTable, capfd):
        _ast = parse(
            """
            lambda: 1
        """
        )

        with mock.patch("sys.exit") as _exit:
            RootContext(_ast)

        output, _ = capfd.readouterr()
        assert "top-level lambdas must be named" in output
        assert _exit.called

    def test_lambda_named(self, parse, RootSymbolTable, capfd):
        # Normal
        _ast = parse(
            """
            x = lambda a, b: 1
            y: type = lambda: 1
            z = lambda *args: 1
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("x", ["a", "b"], None, None),
            Func("y", [], None, None),
            Func("z", [], "args", None),
        )

        # Bad: Multiple LHS and RHS
        _ast = parse(
            """
            x, y = lambda a, b: 1, 2
        """
        )

        with mock.patch("sys.exit") as _exit:
            RootContext(_ast)

        output, _ = capfd.readouterr()
        assert "lambda assignment must be one-to-one" in output
        assert _exit.called

        # Bad: Multiple LHS
        _ast = parse(
            """
            x, y = lambda a: 1
        """
        )

        with mock.patch("sys.exit") as _exit:
            RootContext(_ast)

        output, _ = capfd.readouterr()
        assert "lambda assignment must be one-to-one" in output
        assert _exit.called

        # Bad: Multiple RHS
        _ast = parse(
            """
            x = lambda a, b: 1, 2
        """
        )

        with mock.patch("sys.exit") as _exit:
            RootContext(_ast)

        output, _ = capfd.readouterr()
        assert "lambda assignment must be one-to-one" in output
        assert _exit.called


class TestRootContext_If:
    def test_if(self, parse, RootSymbolTable):
        _ast = parse(
            """
            if some_variable == "some value":
                def fn(a):
                    pass

                global_var = 40
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn", ["a"], None, None),
            Name("global_var"),
        )

    def test_nested_if(self, parse, RootSymbolTable):
        _ast = parse(
            """
            if some_variable == "some value":
                if some_other_variable == "some other value":
                    def fn(a):
                        pass

                    global_var = 40
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn", ["a"], None, None),
            Name("global_var"),
        )


class TestRootContext_For:
    def test_for(self, parse, RootSymbolTable):
        _ast = parse(
            """
            for i in ITER:
                def fn(a):
                    pass

                global_var = 40
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn", ["a"], None, None),
            Name("global_var"),
        )

    def test_nested_for(self, parse, RootSymbolTable):
        _ast = parse(
            """
            for iter in ITER_OF_ITERS:
                for i in iter:
                    def fn(a):
                        pass

                    global_var = 40
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn", ["a"], None, None),
            Name("global_var"),
        )

    def test_async_for(self, parse, RootSymbolTable):
        _ast = parse(
            """
            async for i in ITER:
                def fn(a):
                    pass

                global_var = 40
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn", ["a"], None, None),
            Name("global_var"),
        )


class TestRootContext_While:
    def test_while(self, parse, RootSymbolTable):
        _ast = parse(
            """
            while True:
                def fn(a):
                    pass

                global_var = 40
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn", ["a"], None, None),
            Name("global_var"),
        )

    def test_nested_while(self, parse, RootSymbolTable):
        _ast = parse(
            """
            while True:
                while True:
                    def fn(a):
                        pass

                    global_var = 40
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn", ["a"], None, None),
            Name("global_var"),
        )


class TestRootContext_Try:
    def test_try(self, parse, RootSymbolTable):
        _ast = parse(
            """
            try:
                global_a = 1
            except TypeError:
                global_b = 2
            except:
                global_c = 3
            else:
                global_d = 4
            finally:
                global_e = 5
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Name("global_a"),
            Name("global_b"),
            Name("global_c"),
            Name("global_d"),
            Name("global_e"),
        )

    def test_nested_try(self, parse, RootSymbolTable):
        _ast = parse(
            """
            try:
                try:
                    global_a = 1
                except TypeError:
                    global_b = 2
                except:
                    global_c = 3
                else:
                    global_d = 4
                finally:
                    global_e = 5
            finally:
                global_z = 26
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Name("global_a"),
            Name("global_b"),
            Name("global_c"),
            Name("global_d"),
            Name("global_e"),
            Name("global_z"),
        )


class TestRootContext_With:
    def test_with(self, parse, RootSymbolTable):
        _ast = parse(
            """
            with context_manager() as ctx:
                def fn(a):
                    pass

                global_var = 40
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn", ["a"], None, None),
            Name("global_var"),
        )

    def test_nested_with(self, parse, RootSymbolTable):
        _ast = parse(
            """
            with context_manager() as ctx:
                with ctx() as inner_ctx:
                    def fn(a):
                        pass

                    global_var = 40
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn", ["a"], None, None),
            Name("global_var"),
        )

    def test_async_with(self, parse, RootSymbolTable):
        _ast = parse(
            """
            async with context_manager() as ctx:
                def fn(a):
                    pass

                global_var = 40
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Func("fn", ["a"], None, None),
            Name("global_var"),
        )


class TestRootContext_ClassDefs:
    def test_class_def(self, parse, RootSymbolTable):
        _ast = parse(
            """
            class TopLevel:
                def __init__(self, a):
                    self.a = a

                def method(self):
                    var = "i shouldn't be in the root context!"
                    return var + " see, i'm not!"
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Class("TopLevel", None, None, None),
        )


class TestRootContext_Ignored:
    def test_class_def(self, parse, RootSymbolTable, capfd):
        _ast = parse(
            """
            '''
            Docstring at module level.
            '''
        """
        )
        _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable()

        output, _ = capfd.readouterr()
        assert output == ""


class TestRootContext:

    # Longer "end-to-end" style test -- i.e. context over a whole module

    def test_module(self, parse, RootSymbolTable, capfd):
        _ast = parse(
            """
            import a_module

            from another_module import some_function

            def my_function(a, b):
                sum = a + b
                return sum * sum

            class MyClass:
                def __init__(self):
                    self.data = list()

                def get_data(self):
                    return self.data
        """
        )

        with mock.patch("sys.exit") as _exit:
            _ctx = RootContext(_ast)

        assert _ctx.parent is None
        assert _ctx.symbol_table == RootSymbolTable(
            Import("a_module", "a_module"),
            Import("some_function", "another_module.some_function"),
            Func("my_function", ["a", "b"], None, None),
            Class("MyClass", None, None, None),
        )

        # NOTE `a_module` and `another_module` don't exist
        assert _exit.call_count == 2

        output, _ = capfd.readouterr()
        assert "unable to find module 'a_module'" in output
        assert "unable to find module 'another_module'" in output



"""Rattr quasi-end-to-end test #1."""

import os.path as path
from math import pi


def a_normal_function(arg_one):
    if arg_one.attr > 100:
        raise ValueError("That is too high!")
    return arg_one.attr / 100


def update_attribute(obj, new_value, config=None):
    if config.get("debug"):
        print("Updating .target on", obj.name)

    obj.target = new_value


@circle_condition
def area(circle):
    return pi * circle.radius * circle.radius


@datatype_condition
def is_a_foorbar(datatype, unused_arg):
    filepath = datatype.target_file

    if not path.isfile(filepath):
        return False

    return datatype.is_foo and datatype.is_bar


@assignee_condition
def make_an_assignment(assignee, value):
    assignee.attr = value



"""Assert that there are no assignment to names that identify imports."""

import ast
from typing import List, Union

from rattr.analyser.base import Assertor
from rattr.analyser.types import AnyAssign, AnyFunctionDef, Comprehension
from rattr.analyser.util import (
    get_fullname,
    get_function_def_args,
    has_annotation,
    unravel_names,
)


def get_lhs_names(lhs: AnyAssign) -> List[str]:
    """Return the names in the lhs of the given assignemnt."""
    if hasattr(lhs, "targets"):
        targets = lhs.targets
    else:
        targets = [lhs.target]

    return [n for t in targets for n in unravel_names(t, get_name=get_fullname)]


class ImportClobberingAssertor(Assertor):
    def __clobbered(self, name: str, node: ast.AST) -> None:
        self.failed(f"redefinition of imported name '{name}'", node)

    def __deleted(self, name: str, node: ast.AST) -> None:
        self.failed(f"attempt to delete imported name '{name}'", node)

    def visit_Assign(self, node: ast.Assign) -> None:
        for name in filter(self.context.is_import, get_lhs_names(node)):
            self.__clobbered(name, node)

        return super().generic_visit(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        for name in filter(self.context.is_import, get_lhs_names(node)):
            self.__clobbered(name, node)

        return super().generic_visit(node)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        for name in filter(self.context.is_import, get_lhs_names(node)):
            self.__clobbered(name, node)

        return super().generic_visit(node)

    def visit_Delete(self, node: ast.Delete) -> None:
        for name in filter(self.context.is_import, get_lhs_names(node)):
            self.__deleted(name, node)

        return super().generic_visit(node)

    def visit_FunctionDef(self, node: AnyFunctionDef) -> None:
        if has_annotation("rattr_ignore", node):
            return

        if has_annotation("rattr_results", node):
            return

        if self.context.is_import(node.name):
            self.__clobbered(node.name, node)
            return

        self.check_arguments(node)

        return super().generic_visit(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        return self.visit_FunctionDef(node)

    def visit_Lambda(self, node: ast.Lambda) -> None:
        self.check_arguments(node)
        return super().generic_visit(node)

    def check_arguments(self, node: AnyFunctionDef) -> None:
        args, vararg, kwarg = get_function_def_args(node)
        names = {a for a in (*args, vararg, kwarg) if a is not None}

        for name in filter(lambda n: self.context.is_import(n), names):
            self.__clobbered(name, node)

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        if has_annotation("rattr_ignore", node):
            return

        if has_annotation("rattr_results", node):
            return

        if self.context.is_import(node.name):
            self.__clobbered(node.name, node)
            return

        return super().generic_visit(node)

    def visit_For(self, node: Union[ast.For, ast.AsyncFor]) -> None:
        names = [n for n in unravel_names(node.target)]

        for name in filter(lambda n: self.context.is_import(n), names):
            self.__clobbered(name, node)

        return super().generic_visit(node)

    def visit_AsyncFor(self, node: ast.AsyncFor) -> None:
        return self.visit_For(node)

    def visit_With(self, node: Union[ast.With, ast.AsyncWith]) -> None:
        for item in node.items:
            if item.optional_vars is None:
                continue

            names = [n for n in unravel_names(item.optional_vars)]

            for name in filter(lambda n: self.context.is_import(n), names):
                self.__clobbered(name, node)

        return super().generic_visit(node)

    def visit_AsyncWith(self, node: ast.AsyncWith) -> None:
        return self.visit_With(node)

    def check_comprehension(self, node: Comprehension) -> None:
        for generator in node.generators:
            names = [n for n in unravel_names(generator.target)]

            for name in filter(lambda n: self.context.is_import(n), names):
                self.__clobbered(name, node)

        return super().generic_visit(node)

    def visit_ListComp(self, node: ast.ListComp) -> None:
        self.check_comprehension(node)

    def visit_SetComp(self, node: ast.SetComp) -> None:
        self.check_comprehension(node)

    def visit_GeneratorExp(self, node: ast.GeneratorExp) -> None:
        self.check_comprehension(node)

    def visit_DictComp(self, node: ast.DictComp) -> None:
        self.check_comprehension(node)



"""Symbols for use in context and intermediate results.

Symbol hash and equivalence only holds when the symbols are within the same
context.

Symbol `name` is the Python `_identifier` for the given symbol.

"""

from dataclasses import dataclass
from importlib.machinery import ModuleSpec
from importlib.util import find_spec
from itertools import accumulate, chain, filterfalse
from typing import Dict, List, Optional, Tuple, Union

from rattr import config


@dataclass
class Symbol:
    name: str

    def _is(self, symbol_type) -> bool:
        return isinstance(self, symbol_type)

    def __hash__(self) -> int:
        return hash(repr(self))


@dataclass
class Name(Symbol):
    basename: Optional[str] = None

    def __post_init__(self) -> None:
        if self.basename is None:
            self.basename = self.name

    def __hash__(self) -> int:
        return hash(repr(self))


@dataclass
class Builtin(Symbol):
    has_affect: bool

    def __hash__(self) -> int:
        return hash(repr(self))


@dataclass
class Import(Symbol):
    qualified_name: Optional[str] = None
    module_name: Optional[str] = None
    module_spec: Optional[ModuleSpec] = None

    def __post_init__(self) -> None:
        if self.qualified_name is None:
            self.qualified_name = self.name

        self.module_name, self.module_spec = get_module_name_and_spec(
            self.qualified_name
        )

    def __hash__(self) -> int:
        return hash(repr(self))


@dataclass
class Func(Symbol):
    args: List[str]
    vararg: Optional[str]
    kwarg: Optional[str]
    is_async: bool = False
    defined_in: Optional[str] = None

    def __post_init__(self) -> None:
        self.defined_in = config.current_file

    def __hash__(self) -> int:
        return hash(repr(self))


@dataclass
class Class(Symbol):
    """Represent the class itself and the target of `<class>.__call__()`.

    NOTE
        If no initialiser is found; args, vararg, and kwarg will all be None
        If an initialiser is found; args will be an initialised list (could be
        empty), but vararg and kwarg may-or-may-not be None

    """

    args: Optional[List[str]]
    vararg: Optional[str]
    kwarg: Optional[str]
    defined_in: Optional[str] = None

    def __post_init__(self) -> None:
        self.defined_in = config.current_file

    def __hash__(self) -> int:
        return hash(repr(self))


CallTarget = Union[
    Func,
    Class,
    Builtin,
    Import,
]


@dataclass
class Call(Symbol):
    args: List[str]
    kwargs: Dict[str, str]
    target: Optional[Symbol] = None

    def __hash__(self) -> int:
        return hash(repr(self))


def parse_name(name: str) -> Name:
    return Name(name, name.replace("*", "").split(".")[0])


def parse_call(
    callee: str,
    args: Tuple[List[str], Dict[str, str]],
    target: Optional[CallTarget] = None,
) -> Call:
    if not callee.endswith("()"):
        callee += "()"

    return Call(callee, *args, target)


def get_possible_module_names(name: str) -> List[str]:
    """Return the possible modules for the given name, in relevance order.

    >>> get_possible_module_names("a.b.c.d")
    ["a.b.c.d", "a.b.c", "a.b", "a"]

    >>> get_possible_module_names("os.path.join")
    ["os.path.join", "os.path", "os"]

    Inline comments show the `get_possible_module_names("a.b.c.d")` as a worked
    example.

    """
    # IN: "a.b.c.d"
    # "a", "b", "c", "d"
    parts = name.split(".")

    # ("a", "."), ("b", "."), ("c", "."), ("d", ".")
    dotted_parts = chain.from_iterable(zip(parts, ["."] * len(parts)))

    # "a", "a.", "a.b", "a.b.", "a.b.c", ...
    possible = accumulate(dotted_parts)

    # "a", "a.b", "a.b.c", "a.b.c.d"
    well_formed = filterfalse(lambda p: p.endswith("."), possible)

    # "a.b.c.d", "a.b.c", "a.b", "a"
    ordered = reversed(list(well_formed))

    return list(ordered)


def get_module_name_and_spec(
    name: str,
) -> Union[Tuple[str, ModuleSpec], Tuple[None, None]]:
    """Return the `ModuleSpec` for an imported name.

    If `name` is a module (e.g. `math`) then the spec of the module will be
    returned.

    >>> get_module_name_and_spec("math")
    "math", ModuleSpec(name='math', loader=... origin='built-in')

    If `name` is the fully qualified name of a name within a module then the
    spec of the containing module will be given (e.g. spec of "math" for
    the name "math.pi").

    >>> get_module_name_and_spec("os.path.join")
    "os.path", ModuleSpec(name='posixpath', loader=..., origin=...)

    """
    module_spec_pairs: List[Tuple[str, ModuleSpec]] = list()

    # Filter for valid modles, in acceptance order
    for module in get_possible_module_names(name):
        if module.endswith("."):
            continue

        try:
            spec = find_spec(module)
        except (AttributeError, ModuleNotFoundError, ValueError):
            spec = None

        if spec is None:
            continue

        module_spec_pairs.append((module, spec))

    if len(module_spec_pairs) == 0:
        return None, None

    return module_spec_pairs[0]



from unittest import mock

from rattr import error
from rattr.analyser.util import enter_file


class TestError:
    def test_warning(self, capfd):
        with mock.patch("sys.exit") as _exit:
            error.warning("culprit doesn't matter")

        output, _ = capfd.readouterr()

        assert "warning" in output
        assert not _exit.called

    def test_error(self, capfd):
        with mock.patch("sys.exit") as _exit:
            error.error("culprit doesn't matter")

        output, _ = capfd.readouterr()

        assert "error" in output
        assert not _exit.called

    def test_fatal(self, capfd):
        with mock.patch("sys.exit") as _exit:
            error.fatal("culprit doesn't matter")

        output, _ = capfd.readouterr()

        assert "fatal" in output
        assert _exit.called and _exit.call_count == 1


class TestError_Util:
    def test_get_file_and_line_info(self, parse):
        _ast = parse(
            """
            # not on line 1!
            # not line 2,
            x = 4
            # but third time's the charm
        """
        )
        culprit = _ast.body[0]

        file = "some_file_name.py"
        line_info = "\033[1mline {}:{}: \033[0m".format(
            culprit.lineno, culprit.col_offset
        )
        file_info = "\033[1m{}: \033[0m".format(file)

        # No file, no culprit
        # No file, w/ culprit
        with enter_file(None):
            assert error.get_file_and_line_info(None) == ("", "")
            assert error.get_file_and_line_info(culprit) == ("", line_info)

        # File, no culprit
        # File, w/ culprit
        with enter_file(file):
            assert error.get_file_and_line_info(None) == ("", "")
            assert error.get_file_and_line_info(culprit) == (file_info, line_info)

    def test_split_path(self):
        tests = {
            "": [""],
            # Absolute
            "/": [""],
            "/a": ["", "a"],
            "/a/b": ["", "a", "b"],
            "/a/b/c": ["", "a", "b", "c"],
            # Relative
            ".": ["."],
            # Relative, implicit
            "a": [".", "a"],
            "a/b": [".", "a", "b"],
            "a/b/c": [".", "a", "b", "c"],
            # Relative, explicit
            "./a": [".", "a"],
            "./a/b": [".", "a", "b"],
            "./a/b/c": [".", "a", "b", "c"],
            # Relative to home
            "~": ["~"],
            "~/a": ["~", "a"],
            "~/a/b": ["~", "a", "b"],
            "~/a/b/c": ["~", "a", "b", "c"],
        }

        for test_case, expected in tests.items():
            assert error.split_path(test_case) == expected

    def test_format_path(self):
        assert error.format_path(None) is None
        assert error.format_path("") == ""

        short_paths = [
            "/i/am/short",
            "i/am/short",
            "./i/am/short",
            "~/i/am/short",
        ]

        for path, expected in zip(short_paths, short_paths):
            assert error.format_path(path) == expected

        long_paths = {
            "/fill/fill/fill/i/am/a/long/path": "/.../a/long/path",
            "fill/fill/fill/i/am/a/long/path": "./.../a/long/path",
            "./fill/fill/fill/i/am/a/long/path": "./.../a/long/path",
            "~/fill/fill/fill/i/am/a/long/path": "~/.../a/long/path",
        }

        for path, expected in long_paths.items():
            assert error.format_path(path) == expected



import pytest

from rattr.analyser.context import Call, Class, Func, Name, RootContext
from rattr.analyser.file import FileAnalyser


class TestClassAnalyser:
    def test_class_attributes(self, parse, RootSymbolTable):
        _ast = parse(
            """
            class Numbers(NotARealEnum):
                One = 1
                Two = 2
                Three = 3
                Four, Five = 4, 5
        """
        )
        file_analyser = FileAnalyser(_ast, RootContext(_ast))
        file_analyser.analyse()

        assert file_analyser.context.symbol_table == RootSymbolTable(
            Class("Numbers", None, None, None),
            Name("Numbers.One", "Numbers"),
            Name("Numbers.Two", "Numbers"),
            Name("Numbers.Three", "Numbers"),
            Name("Numbers.Four", "Numbers"),
            Name("Numbers.Five", "Numbers"),
        )

    def test_class_initialiser(self, parse):
        _ast = parse(
            """
            class SomeClass:
                def __init__(self, some_arg):
                    self.whatever = some_arg.whatever
                    some_arg.good_number = 5    # 5 is a good number
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Class("SomeClass", ["self", "some_arg"], None, None): {
                "sets": {
                    Name("self.whatever", "self"),
                    Name("some_arg.good_number", "some_arg"),
                },
                "gets": {
                    Name("some_arg.whatever", "some_arg"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

    def test_class_initialiser_integration(self, parse):
        # Func -> __init__
        _ast = parse(
            """
            def func(arg):
                return SomeClass(arg)

            class SomeClass:
                def __init__(self, some_arg):
                    self.whatever = some_arg.whatever
                    some_arg.good_number = 5    # 5 is a good number
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        func = Func("func", ["arg"], None, None)
        cls = Class("SomeClass", ["self", "some_arg"], None, None)
        expected = {
            func: {
                "sets": set(),
                "gets": {Name("arg")},
                "dels": set(),
                "calls": {Call("SomeClass()", ["@ReturnValue", "arg"], {}, target=cls)},
            },
            cls: {
                "sets": {
                    Name("self.whatever", "self"),
                    Name("some_arg.good_number", "some_arg"),
                },
                "gets": {
                    Name("some_arg.whatever", "some_arg"),
                },
                "dels": set(),
                "calls": set(),
            },
        }

        # HACK This has to be a repr, can't just compare, idk why
        print(results)
        assert repr(results._file_ir) == repr(expected)

        # __init__ -> Func
        _ast = parse(
            """
            def func(fn_arg):
                return fn_arg.some_attr

            class SomeClass:
                def __init__(self, some_arg):
                    self.a_thing = func(some_arg)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        func = Func("func", ["fn_arg"], None, None)
        cls = Class("SomeClass", ["self", "some_arg"], None, None)
        expected = {
            func: {
                "sets": set(),
                "gets": {Name("fn_arg.some_attr", "fn_arg")},
                "dels": set(),
                "calls": set(),
            },
            cls: {
                "sets": {
                    Name("self.a_thing", "self"),
                },
                "gets": {
                    Name("some_arg"),
                },
                "dels": set(),
                "calls": {Call("func()", ["some_arg"], {}, target=func)},
            },
        }

        assert results == expected

    def test_staticmethod(self, parse):
        # Simple
        _ast = parse(
            """
            class SomeClass:
                @staticmethod
                def method(a, b):
                    a.attr = b.attr
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        method = Func("SomeClass.method", ["a", "b"], None, None)
        expected = {
            method: {
                "sets": {
                    Name("a.attr", "a"),
                },
                "gets": {Name("b.attr", "b")},
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

        # Static method in "mixed" class
        _ast = parse(
            """
            class SomeClass:
                def __init__(self, arg):
                    self.attr = arg

                @staticmethod
                def method(a, b):
                    a.attr = b.attr
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        cls = Class("SomeClass", ["self", "arg"], None, None)
        method = Func("SomeClass.method", ["a", "b"], None, None)
        expected = {
            cls: {
                "sets": {Name("self.attr", "self")},
                "gets": {Name("arg")},
                "dels": set(),
                "calls": set(),
            },
            method: {
                "sets": {
                    Name("a.attr", "a"),
                },
                "gets": {Name("b.attr", "b")},
                "dels": set(),
                "calls": set(),
            },
        }

        assert results == expected

    def test_class_full(self, parse):
        # Example 1
        _ast = parse(
            """
            class FibIter:
                N  = 0
                N_ = 1

                @staticmethod
                def next() -> int:
                    last = FibIter.N

                    FibIter.N  = FibIter.N_
                    FibIter.N_ = FibIter.N_ + last

                    return FibIter.N
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        method = Func("FibIter.next", [], None, None)
        expected = {
            method: {
                "sets": {
                    Name("last"),
                    Name("FibIter.N", "FibIter"),
                    Name("FibIter.N_", "FibIter"),
                },
                "gets": {
                    Name("last"),
                    Name("FibIter.N", "FibIter"),
                    Name("FibIter.N_", "FibIter"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

    def test_enum(self, parse):
        # Explicit import
        _ast = parse(
            """
            from enum import Enum

            class MyEnum(Enum):
                LHS = "RHS"
        """
        )
        _ctx = RootContext(_ast).expand_starred_imports()
        results = FileAnalyser(_ast, _ctx).analyse()

        enum = Class("MyEnum", ["self", "_id"], None, None)
        expected = {
            enum: {
                "sets": set(),
                "gets": {
                    Name("MyEnum.LHS", "MyEnum"),
                },
                "calls": set(),
                "dels": set(),
            }
        }

        assert results == expected

        # Module import
        _ast = parse(
            """
            import enum

            class MyEnum(enum.Enum):
                LHS = "RHS"
        """
        )
        _ctx = RootContext(_ast).expand_starred_imports()
        results = FileAnalyser(_ast, _ctx).analyse()

        enum = Class("MyEnum", ["self", "_id"], None, None)
        expected = {
            enum: {
                "sets": set(),
                "gets": {
                    Name("MyEnum.LHS", "MyEnum"),
                },
                "calls": set(),
                "dels": set(),
            }
        }

        assert results == expected

        # Implicit import
        _ast = parse(
            """
            from enum import *

            class MyEnum(Enum):
                LHS = "RHS"
        """
        )
        _ctx = RootContext(_ast).expand_starred_imports()
        results = FileAnalyser(_ast, _ctx).analyse()

        enum = Class("MyEnum", ["self", "_id"], None, None)
        expected = {
            enum: {
                "sets": set(),
                "gets": {
                    Name("MyEnum.LHS", "MyEnum"),
                },
                "calls": set(),
                "dels": set(),
            }
        }

        assert results == expected

    def test_enum_call(self, parse, constant):
        _ast = parse(
            """
            from enum import Enum

            class MyEnum(Enum):
                ONE = "one"
                TWO = "two"
                THREE = "three"

            def get_one():
                # actually gets all!
                return MyEnum("one")

            def get_none():
                return 4
        """
        )
        _ctx = RootContext(_ast).expand_starred_imports()
        results = FileAnalyser(_ast, _ctx).analyse()

        enum = Class("MyEnum", ["self", "_id"], None, None)
        get_one = Func("get_one", [], None, None)
        get_none = Func("get_none", [], None, None)
        expected = {
            enum: {
                "sets": set(),
                "gets": {
                    Name("MyEnum.ONE", "MyEnum"),
                    Name("MyEnum.TWO", "MyEnum"),
                    Name("MyEnum.THREE", "MyEnum"),
                },
                "calls": set(),
                "dels": set(),
            },
            get_one: {
                "sets": set(),
                "gets": set(),
                "calls": {
                    Call("MyEnum()", ["@ReturnValue", constant("Str")], {}, target=enum)
                },
                "dels": set(),
            },
            get_none: {
                "sets": set(),
                "gets": set(),
                "calls": set(),
                "dels": set(),
            },
        }

        print(results)
        assert results == expected

    @pytest.mark.py_3_8_plus()
    def test_walrus(self, parse, RootSymbolTable):
        # "Normal"
        _ast = parse(
            """
            class IAmTheWalrus:
                cls_attr_a = (cls_attr_b := 0)  # Yuck

                def __init__(self):
                    inst_attr_a = (inst_attr_b := 0)  # Just as yuck
            """
        )
        file_analyser = FileAnalyser(_ast, RootContext(_ast))
        results = file_analyser.analyse()

        wally = Class("IAmTheWalrus", ["self"], None, None)
        cls_attr_a = Name("IAmTheWalrus.cls_attr_a", "IAmTheWalrus")
        cls_attr_b = Name("IAmTheWalrus.cls_attr_b", "IAmTheWalrus")

        assert file_analyser.context.symbol_table == RootSymbolTable(
            wally, cls_attr_a, cls_attr_b
        )

        expected = {
            wally: {
                "sets": {
                    Name("inst_attr_a"),
                    Name("inst_attr_b"),
                },
                "gets": set(),
                "dels": set(),
                "calls": set(),
            }
        }

        print(results)
        assert results == expected

        # Tuple'd
        _ast = parse(
            """
            class IAmTheWalrus:
                cls_attr_a = (a, cls_attr_b := 0)

                def __init__(self):
                    inst_attr_a = (a, inst_attr_b := 0)
            """
        )
        file_analyser = FileAnalyser(_ast, RootContext(_ast))
        results = file_analyser.analyse()

        wally = Class("IAmTheWalrus", ["self"], None, None)
        cls_attr_a = Name("IAmTheWalrus.cls_attr_a", "IAmTheWalrus")
        cls_attr_b = Name("IAmTheWalrus.cls_attr_b", "IAmTheWalrus")

        assert file_analyser.context.symbol_table == RootSymbolTable(
            wally, cls_attr_a, cls_attr_b
        )

        expected = {
            wally: {
                "sets": {
                    Name("inst_attr_a"),
                    Name("inst_attr_b"),
                },
                "gets": {
                    Name("a"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        print(results)
        assert results == expected

    @pytest.mark.py_3_8_plus()
    def test_walrus_method(self, parse):
        _ast = parse(
            """
            class IAmTheWalrus:
                def contrived(self):
                    this_is = (a_very_contrived := self.operation)

                @staticmethod
                def more_contrived():
                    if (alias := this_is_just_so_long_i_will_alias_it):
                        return alias
            """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        static = Func("IAmTheWalrus.more_contrived", [], None, None, False, None)
        expected = {
            static: {
                "sets": {
                    Name("alias"),
                },
                "gets": {
                    Name("this_is_just_so_long_i_will_alias_it"),
                    Name("alias"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        print(results)
        assert results == expected



import pathlib

from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent.resolve()

# Copy description from README.md
with open(HERE / "README.md") as f:
    DESCRIPTION = f.read()

# Copy requirements from requirements.txt
with open(HERE / "requirements.txt") as f:
    INSTALL_REQUIRES = [line.strip() for line in f.readlines()]


setup(
    name="rattr",
    use_scm_version={
        "write_to": "rattr/_version.py",
        "write_to_template": (
            'version = "{version}"  # this should be overwritten by setuptools_scm\n'
        ),
    },
    setup_requires=["setuptools_scm"],
    author="Suade Labs",
    packages=find_packages(),
    description="Rattr rats on your attrs.",
    long_description=DESCRIPTION,
    long_description_content_type="text/markdown",
    install_requires=INSTALL_REQUIRES,
    extras_require={
        "dev": [
            "black==22.1.0",
            "click==8.0.2",
            "flake8-bugbear==22.1.11",
            "flake8==4.0.1",
            "flask==2.0.3",
            "isort==5.10.1",
            "pytest==7.0.1",
        ],
    },
    entry_points={"console_scripts": ["rattr = rattr.__main__:entry_point"]},
    keywords="automation linting type-checking attributes rats",
    url="https://github.com/SuadeLabs/ratter",
    license="MIT",
    python_requires=">=3.7",
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Quality Assurance",
    ],
)



from rattr.analyser.context import Builtin, Call, Func, Name, RootContext
from rattr.analyser.file import FileAnalyser


class TestRegression:
    def test_highly_nested(self, parse):
        _ast = parse(
            """
            def getter(a):
                *a.b[0]().c

            def setter(a):
                *a.b[0]().c = "a value"
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("getter", ["a"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("*a.b[]().c", "a"),
                },
                "sets": set(),
            },
            Func("setter", ["a"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": set(),
                "sets": {
                    Name("*a.b[]().c", "a"),
                },
            },
        }

        assert results == expected

    def test_tuple_assignment(self, parse):
        _ast = parse(
            """
            def getter(a):
                x, y = a.attr_a, a.attr_b

            def setter(a):
                a.attr_a, a.attr_b = x, y
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("getter", ["a"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("a.attr_a", "a"),
                    Name("a.attr_b", "a"),
                },
                "sets": {
                    Name("x"),
                    Name("y"),
                },
            },
            Func("setter", ["a"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("x"),
                    Name("y"),
                },
                "sets": {
                    Name("a.attr_a", "a"),
                    Name("a.attr_b", "a"),
                },
            },
        }

        assert results == expected

    def test_operation_on_anonymous_return_value(self, parse, constant):
        _ast = parse(
            """
            def test_1(a):
                # i.e. b.c and b.d are of a type that implements __add__
                a = max((b.c + b.d).method(), 0)

            def test_2(a):
                return (-b).c
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        max_symbol = Builtin("max", has_affect=False)
        expected = {
            Func("test_1", ["a"], None, None): {
                "calls": {
                    Call("@BinOp.method()", [], {}, None),
                    Call("max()", ["@BinOp.method()", constant("Num")], {}, max_symbol),
                },
                "dels": set(),
                "gets": set(),
                "sets": {
                    Name("a"),
                },
            },
            Func("test_2", ["a"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("b"),
                    Name("@UnaryOp.c", "@UnaryOp"),
                },
                "sets": set(),
            },
        }

        print(results)
        assert results == expected

    def test_resolve_getattr_name(self, parse):
        _ast = parse(
            """
            def act(on):
                return on.attr

            def a_func(arg):
                return act(getattr(arg, "attr"))
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        fn_act = Func("act", ["on"], None, None)
        fn_a = Func("a_func", ["arg"], None, None)
        expected = {
            fn_act: {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("on.attr", "on"),
                },
                "sets": set(),
            },
            fn_a: {
                "calls": {Call("act()", ["arg.attr"], {}, target=fn_act)},
                "dels": set(),
                "gets": {
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

    def test_higher_order_functions(self, parse, capfd):
        # NOTE
        #   Higher-order functions will not give the "correct" results as they
        #   can't realistically be statically resolved, however, they should
        #   not cause a crash or a fatal error.

        # Procedural parameter
        _ast = parse(
            """
            def bad_map(f, target):
                # like map but... bad
                return f(target)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        bad_map = Func("bad_map", ["f", "target"], None, None)
        expected = {
            bad_map: {
                "sets": set(),
                "gets": {Name("target")},
                "dels": set(),
                "calls": {Call("f()", ["target"], {}, target=Name("f"))},
            }
        }

        assert results == expected

        output, _ = capfd.readouterr()
        assert "likely a procedural parameter" in output

        # Procedural return value
        _ast = parse(
            """
            def wrapper():
                def inner():
                    return "some value"
                return inner

            def actor():
                return wrapper()()
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        wrapper = Func("wrapper", [], None, None)
        actor = Func("actor", [], None, None)
        expected = {
            wrapper: {
                "sets": set(),
                "gets": {Name("inner")},
                "dels": set(),
                "calls": set(),
            },
            actor: {
                "sets": set(),
                "gets": set(),
                "dels": set(),
                "calls": {Call("wrapper()()", [], {}, target=wrapper)},
            },
        }

        assert results == expected

        output, _ = capfd.readouterr()
        assert "unable to resolve call result of call" in output

        # Local function
        # Technically not higher-order, but w/e
        _ast = parse(
            """
            def fn_a(arg):
                return arg.a

            def fn_b(arg):
                return arg.b

            def bad(argument) -> int:
                accumulator = 0

                for f in [fn_a, fn_b]:
                    accumulator += f(argument)

                return accumulator
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        fn_a = Func("fn_a", ["arg"], None, None)
        fn_b = Func("fn_b", ["arg"], None, None)
        bad = Func("bad", ["argument"], None, None)
        expected = {
            fn_a: {
                "sets": set(),
                "gets": {Name("arg.a", "arg")},
                "dels": set(),
                "calls": set(),
            },
            fn_b: {
                "sets": set(),
                "gets": {Name("arg.b", "arg")},
                "dels": set(),
                "calls": set(),
            },
            bad: {
                "sets": {
                    Name("accumulator"),
                    Name("f"),
                },
                "gets": {
                    Name("accumulator"),
                    Name("fn_a"),
                    Name("fn_b"),
                    Name("argument"),
                },
                "dels": set(),
                "calls": {Call("f()", ["argument"], {}, target=Name("f"))},
            },
        }

        assert results == expected

        output, _ = capfd.readouterr()
        assert "likely a procedural parameter" in output

    def test_regression_generator_iterable_is_literal(self, parse):
        # List of many
        _ast = parse(
            """
            def fn():
                list_of_tuples = [
                    ("first", [
                        e.whatever for e in [thing_one, thing_two]
                    ])
                ]
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("fn", [], None, None): {
                "sets": {
                    Name("list_of_tuples"),
                },
                "gets": {
                    Name("thing_one"),
                    Name("thing_two"),
                    Name("e"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

        # List of one
        _ast = parse(
            """
            def fn():
                list_of_tuples = [
                    ("first", [
                        e.whatever for e in [thing]
                    ])
                ]
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("fn", [], None, None): {
                "sets": {
                    Name("list_of_tuples"),
                },
                "gets": {
                    Name("thing"),
                    Name("e"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

    def test_regression_gets_in_bin_op(self, parse):
        # NOTE in Python 3.8 support / tests for "sets" in BinOp will be needed

        #
        # In Attribute
        #

        # ALWAYS WORKED
        # Test if "arg.first" and "arg.second" are in "gets"
        _ast = parse(
            """
            def fn(arg):
                return (arg.first + arg.second)
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("fn", ["arg"], None, None): {
                "sets": set(),
                "gets": {
                    Name("arg.first", "arg"),
                    Name("arg.second", "arg"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

        # REGRESSION
        # Test if "arg.first" and "arg.second" are in "gets"
        _ast = parse(
            """
            def fn(arg):
                return (arg.first + arg.second).some_property
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("fn", ["arg"], None, None): {
                "sets": set(),
                "gets": {
                    Name("arg.first", "arg"),
                    Name("arg.second", "arg"),
                    Name("@BinOp.some_property", "@BinOp"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

        #
        # In Starred
        #
        _ast = parse(
            """
            def fn(arg):
                return call(*[arg.first + arg.second])
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("fn", ["arg"], None, None): {
                "sets": set(),
                "gets": {
                    Name("arg.first", "arg"),
                    Name("arg.second", "arg"),
                    Name("*@List", "@List"),
                },
                "dels": set(),
                "calls": {
                    Call("call()", ["*@List"], {}, None),
                },
            }
        }

        assert results == expected

        #
        # In Subscript
        #
        _ast = parse(
            """
            def fn(arg):
                return (arg.first + arg.second)["index"]
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("fn", ["arg"], None, None): {
                "sets": set(),
                "gets": {
                    Name("arg.first", "arg"),
                    Name("arg.second", "arg"),
                    Name("@BinOp[]", "@BinOp"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        print(results)
        assert results == expected

    def test_regression_gets_in_method_call(self, parse):
        # ALWAYS WORKED
        _ast = parse(
            """
            def fn(arg):
                return arg.attr
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("fn", ["arg"], None, None): {
                "sets": set(),
                "gets": {
                    Name("arg.attr", "arg"),
                },
                "dels": set(),
                "calls": set(),
            }
        }

        assert results == expected

        # ALWAYS WORKED
        _ast = parse(
            """
            def fn(arg):
                return arg.method()
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("fn", ["arg"], None, None): {
                "sets": set(),
                "gets": set(),
                "dels": set(),
                "calls": {
                    Call("arg.method()", [], {}, Name("arg")),
                },
            }
        }

        print(results)
        assert results == expected

        # REGRESSION 1
        _ast = parse(
            """
            def fn(arg):
                return arg.attr.method()
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("fn", ["arg"], None, None): {
                "sets": set(),
                "gets": {
                    Name("arg.attr", "arg"),
                },
                "dels": set(),
                "calls": {
                    Call("arg.attr.method()", [], {}, Name("arg")),
                },
            }
        }

        print(results)
        assert results == expected

        # REGRESSION 2
        _ast = parse(
            """
            def fn(arg):
                return arg.a.b.c.d.method()
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("fn", ["arg"], None, None): {
                "sets": set(),
                "gets": {
                    Name("arg.a", "arg"),
                    Name("arg.a.b", "arg"),
                    Name("arg.a.b.c", "arg"),
                    Name("arg.a.b.c.d", "arg"),
                },
                "dels": set(),
                "calls": {
                    Call("arg.a.b.c.d.method()", [], {}, Name("arg")),
                },
            }
        }

        print(results)
        assert results == expected



"""Tests shared across multiple analysers."""

from unittest import mock

import pytest

from rattr.analyser.context import RootContext
from rattr.analyser.context.symbol import Call, Func, Name
from rattr.analyser.file import FileAnalyser


class TestAnnotations:
    def test_rattr_ignore(self, parse):
        # FunctionDef
        _ast = parse(
            """
            def a_func(arg):
                return arg.attr

            @rattr_ignore
            def another_func(arg):
                return arg.another_attr
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

        # AsyncFunctionDef
        _ast = parse(
            """
            async def a_func(arg):
                return arg.attr

            @rattr_ignore
            async def another_func(arg):
                return arg.another_attr
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None, is_async=True): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

        # ClassDef
        # AsyncFunctionDef
        _ast = parse(
            """
            async def a_func(arg):
                return arg.attr

            @rattr_ignore
            class SomeClass:
                def __init__(self, a):
                    self.a = a
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None, is_async=True): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
            },
        }

        assert results == expected

    def test_rattr_results(self, parse):
        # FunctionDef
        # Use incorrect results to show override is working
        _ast = parse(
            """
            def a_func(arg):
                return arg.attr

            @rattr_results(gets={"a", "b"})
            def another_func(arg):
                return arg.another_attr
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
            },
            Func("another_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("a"),
                    Name("b"),
                },
                "sets": set(),
            },
        }

        assert results == expected

        # AsyncFunctionDef
        # Use incorrect results to show override is working
        _ast = parse(
            """
            async def a_func(arg):
                return arg.attr

            @rattr_results()
            async def another_func(arg):
                return arg.another_attr
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None, is_async=True): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
            },
            Func("another_func", ["arg"], None, None, is_async=True): {
                "calls": set(),
                "dels": set(),
                "gets": set(),
                "sets": set(),
            },
        }

        assert results == expected

        # ClassDef
        # TODO When classes added

    def test_rattr_results_complex(self, parse):
        # FunctionDef
        # Use incorrect results to show override is working
        _ast = parse(
            """
            def a_func(arg):
                return arg.attr

            @rattr_results(
                gets={"a.attr", "*b.value"},
                calls=[
                    ("fn_a()", (["a", "a.attr"], {"key": "b.key"}))
                ]
            )
            def another_func(arg):
                return arg.another_attr
        """
        )
        results = FileAnalyser(_ast, RootContext(_ast)).analyse()

        expected = {
            Func("a_func", ["arg"], None, None): {
                "calls": set(),
                "dels": set(),
                "gets": {
                    Name("arg.attr", "arg"),
                },
                "sets": set(),
            },
            Func("another_func", ["arg"], None, None): {
                "calls": {Call("fn_a()", ["a", "a.attr"], {"key": "b.key"})},
                "dels": set(),
                "gets": {
                    Name("a.attr", "a"),
                    Name("*b.value", "b"),
                },
                "sets": set(),
            },
        }

        print(results)
        assert results == expected



from dataclasses import dataclass
from typing import List, Set, Union

from rattr.analyser.base import (
    Assertor,
    CustomFunctionAnalyser,
    CustomFunctionHandler,
)
from rattr.analyser.util import MODULE_BLACKLIST_PATTERNS
from rattr.plugins.analysers import DEFAULT_FUNCTION_ANALYSERS as DEFAULTS

Plugin = Union[Assertor, CustomFunctionAnalyser]


@dataclass
class Plugins:
    assertors: List[Assertor]
    analysers: List[CustomFunctionAnalyser]

    _handler: CustomFunctionHandler = None

    @property
    def custom_function_handler(self) -> CustomFunctionHandler:
        if self._handler is None:
            self._handler = CustomFunctionHandler(DEFAULTS, self.analysers)

        return self._handler

    def register(self, plugin: Plugin) -> None:
        """Register the given plugin."""
        if isinstance(plugin, Assertor):
            return self.assertors.append(plugin)

        if isinstance(plugin, CustomFunctionAnalyser):
            return self.analysers.append(plugin)

        raise TypeError

    def blacklist(self, module_or_modules: Union[str, Set[str]]) -> None:
        """Add the given module pattern to the module blacklist."""
        if not isinstance(module_or_modules, (str, set)):
            raise TypeError

        if isinstance(module_or_modules, str):
            modules = {
                module_or_modules,
            }

        if isinstance(module_or_modules, set):
            modules = module_or_modules

        for module in modules:
            MODULE_BLACKLIST_PATTERNS.add(module)


plugins = Plugins(list(), list())



"""Rattr CLI parser util functions."""

from argparse import Namespace
from textwrap import dedent, fill
from typing import Any, Dict


def multi_paragraph_wrap(text: str, width: int = 80) -> str:
    """Return the text dedented and wrapped at 70 characters.

    Assumes that paragraphs are separated by a double newline, with single
    newlines being removed.

    Paragraphs whose lines begin with ">" have their singlelines and relative
    indentation preserved. Furthermore, the string "# noqa" (and any preceding
    whitespace) is removed from the end of the lines, if present, allowing for
    linter complaints to be ignored while preserving formatting for long lines.

    """

    def _preserve(text: str) -> str:
        lines = list()

        for line in text.splitlines():
            if not line.startswith(">"):
                raise SyntaxError("preserved lines must start with '>'")
            else:
                line = line[1:]

            if line.endswith("# noqa"):
                line = line[:-6].rstrip()

            lines.append(line)

        return "\n".join(fill(dedent(line), width) for line in lines)

    def _paragraph(text: str) -> str:
        return fill(text, width)

    paragraphs = list()

    for p in dedent(text).split("\n\n"):
        if p.startswith(">"):
            paragraphs.append(_preserve(p))
        else:
            paragraphs.append(_paragraph(p))

    return "\n\n".join(p for p in paragraphs)


def namespace_to_dict(arguments: Namespace) -> Dict[str, Any]:
    """Return the given arguments as a dictionary."""
    return {k: v for k, v in arguments._get_kwargs()}
